const std = @import("std");
const sdl = @import("sdl2");
const zlm = @import("zlm");
const gl = @import("zgl");
const m = zlm.SpecializeOn(f32);

const Self = @This();

window: sdl.Window,
context: sdl.gl.Context,
buffer: gl.Buffer,
color: [4]f32 = .{ 0, 0, 0, 0 },

// There's a vbo for each shader program
vbo: [max_objects]f32 = .{0.0} ** max_objects,
vbo_index: usize = 0,

// The index of the program is the enum converted to int
textures: [max_objects]Texture = .{undefined} ** max_objects,

const max_objects: usize = 16384;
const quadSize: usize = 9 * 6;

fn glGetProcAddress(p: []const u8, proc: [:0]const u8) ?*const anyopaque {
    _ = p;
    return sdl.c.SDL_GL_GetProcAddress(@ptrCast([*c]const u8, proc));
}

pub fn init() !Self {
    try sdl.init(.{ .video = true, .audio = true, .events = true });
    try sdl.ttf.init();

    try sdl.gl.setAttribute(.{ .context_major_version = 3 });
    try sdl.gl.setAttribute(.{ .context_minor_version = 3 });
    try sdl.gl.setAttribute(.{ .multisamplebuffers = true });
    try sdl.gl.setAttribute(.{ .multisamplesamples = 4 });

    const window = try sdl.createWindow(
        "USG",
        .{ .centered = {} },
        .{ .centered = {} },
        1280,
        720,
        .{ .context = .opengl, .vis = .shown },
    );

    const proc: []const u8 = undefined;
    try gl.loadExtensions(proc, glGetProcAddress);

    const ctx = try sdl.gl.createContext(window);
    sdl.gl.setSwapInterval(.immediate) catch {
        std.debug.print("WARNING: Unable to configure the swap interval.\n", .{});
    };

    // Activate blending and configure it
    gl.enable(.blend);
    gl.blendFunc(.src_alpha, .one_minus_src_alpha);

    var buf = gl.Buffer.gen();
    buf.bind(.array_buffer);

    var renderer = Self{ .window = window, .context = ctx, .buffer = buf };

    // Load the shader programs according to the names in the enum
    _ = renderer.loadProgram("texture").use();

    var vertex_array = gl.VertexArray.gen();
    vertex_array.bind();

    gl.vertexAttribPointer(0, 2, .float, false, @sizeOf(f32) * 9, @sizeOf(f32) * 0);
    gl.enableVertexAttribArray(0);

    gl.vertexAttribPointer(1, 4, .float, false, @sizeOf(f32) * 9, @sizeOf(f32) * 2);
    gl.enableVertexAttribArray(1);

    gl.vertexAttribPointer(2, 2, .float, false, @sizeOf(f32) * 9, @sizeOf(f32) * 6);
    gl.enableVertexAttribArray(2);

    gl.vertexAttribPointer(3, 1, .float, false, @sizeOf(f32) * 9, @sizeOf(f32) * 8);
    gl.enableVertexAttribArray(3);

    gl.clearColor(0.91, 0.85, 0.65, 1.00);

    return renderer;
}

fn loadProgram(self: Self, comptime name: []const u8) gl.Program {
    var program = gl.Program.create();

    const vs = gl.Shader.create(.vertex);
    defer vs.delete();
    const fs = gl.Shader.create(.fragment);
    defer fs.delete();

    vs.source(1, &.{@embedFile("shaders/" ++ name ++ ".vert")});
    fs.source(1, &.{@embedFile("shaders/" ++ name ++ ".frag")});

    vs.compile();
    fs.compile();

    program.attach(vs);
    defer program.detach(vs);
    program.attach(fs);
    defer program.detach(fs);

    program.link();

    // Get the matrix that converts coordinates to 1:1 on the screen from the OpenGL default
    if (program.uniformLocation("mvp")) |mvpLocation| {
        const wsize = self.window.getSize();
        const xunit = @intToFloat(f32, wsize.width);
        const yunit = @intToFloat(f32, wsize.height);

        const ortho = m.Mat4.createOrthogonal(0.0, xunit, yunit, 0.0, 0.0, 100.0);

        program.use();
        gl.uniformMatrix4fv(mvpLocation, false, &.{ortho.fields});
    }

    return program;
}

pub fn render(self: *Self) void {
    // TODO: Submit with SubData instead to not pass the whole buffer
    self.buffer.data(f32, &self.vbo, .static_draw);

    var i: usize = 0;
    const amount = self.vbo_index / quadSize;
    while (i < amount) : (i += 1) {
        gl.bindTexture(self.textures[i].texture, .@"2d");
        gl.drawArrays(.triangles, i * 6, 6);
    }
    self.vbo_index = 0;

    // Clear the vbo, this really shouldn't be necessary as the index is set to zero,
    // but otherwise it leads to bugs
    self.vbo = .{0.0} ** max_objects;

    sdl.gl.swapWindow(self.window);
    gl.clear(.{ .color = true });
}

pub fn deinit(self: *Self) void {
    gl.disableVertexAttribArray(0);
    gl.disableVertexAttribArray(1);
    gl.disableVertexAttribArray(2);
    gl.disableVertexAttribArray(3);
    sdl.quit();
    sdl.ttf.quit();
    self.window.destroy();
    sdl.gl.deleteContext(self.context);
}

pub fn setColorF(self: *Self, r: f32, g: f32, b: f32, a: f32) void {
    self.*.color = .{ r, g, b, a };
}

pub fn setColor(self: *Self, r: i32, g: i32, b: i32, a: i32) void {
    self.*.color = .{
        @intToFloat(f32, r) / 255.0,
        @intToFloat(f32, g) / 255.0,
        @intToFloat(f32, b) / 255.0,
        @intToFloat(f32, a) / 255.0,
    };
}

pub fn drawRectangle(self: *Self, x: i32, y: i32, w: i32, h: i32) void {
    //const indices = [_]u8{ 0, 1, 1, 2, 2, 3, 3, 0 };
    //self.renderRectangle(x, y, w, h, .lines, &indices);
    _ = x;
    _ = y;
    _ = self;
    _ = w;
    _ = h;
}

pub fn fillRectangle(self: *Self, x: i32, y: i32, w: i32, h: i32) void {
    self.fillRectangleEx(x, y, w, h, 0);
}

pub fn fillRectangleEx(self: *Self, x: i32, y: i32, w: i32, h: i32, skew_x: i32) void {
    renderRectangle(self, x, y, w, h, skew_x, 0);
}

pub fn renderTexture(self: *Self, texture: Texture, x: i32, y: i32) void {
    self.textures[self.vbo_index / quadSize] = texture;
    renderRectangle(self, x, y, @intCast(i32, texture.width), @intCast(i32, texture.height), 0, 1);
}

fn renderRectangle(self: *Self, x: i32, y: i32, w: i32, h: i32, skew_x: i32, unit: f32) void {
    var xf = @intToFloat(f32, x);
    var yf = @intToFloat(f32, y);
    var wf = @intToFloat(f32, w);
    var hf = @intToFloat(f32, h);

    const skew_x_offset = @intToFloat(f32, skew_x) * hf / 100;

    const i = self.vbo_index;
    const vertex_data = [_]f32{
        xf + skew_x_offset, yf, // up-left
        self.color[0],      self.color[1],
        self.color[2],      self.color[3],
        0,                  0,
        unit,
        xf + wf + skew_x_offset, yf, // up-right
        self.color[0],           self.color[1],
        self.color[2],           self.color[3],
        1,                       0,
        unit,
        xf - skew_x_offset, yf + hf, // down-left
        self.color[0],      self.color[1],
        self.color[2],      self.color[3],
        0,                  1,
        unit,
        xf + wf - skew_x_offset, yf + hf, // down-right
        self.color[0],           self.color[1],
        self.color[2],           self.color[3],
        1,                       1,
        unit,
        xf + wf + skew_x_offset, yf, // up-right
        self.color[0],           self.color[1],
        self.color[2],           self.color[3],
        1,                       0,
        unit,
        xf - skew_x_offset, yf + hf, // down-left
        self.color[0],      self.color[1],
        self.color[2],      self.color[3],
        0,                  1,
        unit,
    };
    std.mem.copy(f32, self.vbo[i..], &vertex_data);
    self.vbo_index += vertex_data.len;
}

pub const OutputSize = struct { width: c_int, height: c_int };
pub fn getOutputSize(self: Self) OutputSize {
    var wsize = self.window.getSize();
    return OutputSize{
        .width = wsize.width,
        .height = wsize.height,
    };
}

pub const Texture = struct {
    texture: gl.Texture,
    width: usize,
    height: usize,

    pub fn init(data: [*]const u8, width: usize, height: usize) Texture {
        var tex = gl.genTexture();
        //defer gl.Texture.delete(tex);
        gl.bindTexture(tex, .@"2d");

        gl.textureImage2D(.@"2d", 0, .rgba, width, height, .rgba, .unsigned_byte, data);

        gl.texParameter(.@"2d", .wrap_s, .clamp_to_edge);
        gl.texParameter(.@"2d", .wrap_r, .clamp_to_edge);
        gl.texParameter(.@"2d", .min_filter, .linear);
        gl.texParameter(.@"2d", .mag_filter, .linear);

        return Texture{
            .texture = tex,
            .width = width,
            .height = height,
        };
    }

    pub fn fromText(text: [:0]const u8, size: c_int) Texture {
        var font = sdl.ttf.openFont("res/fonts/MarginaliaRegular-8XlZ.ttf", size) catch unreachable;
        defer font.close();

        var surface = font.renderTextBlended(text, sdl.Color.white) catch unreachable;
        defer surface.destroy();

        const width = @intCast(usize, surface.ptr.w);
        const height = @intCast(usize, surface.ptr.h);
        var newsurf = sdl.c.SDL_ConvertSurfaceFormat(surface.ptr, @enumToInt(sdl.PixelFormatEnum.argb8888), 0) orelse unreachable;
        const data = @ptrCast([*]const u8, newsurf.pixels);

        return Texture.init(data, width, height);
    }

    pub fn deinit(self: *Texture) void {
        gl.Texture.delete(self.texture);
    }
};
